@@ -9,13 +9,10 @@
 package SRS::EPP::Proxy;
 
 use MooseX::Singleton;
-use MooseX::Method::Signatures;
-
+use MooseX::Params::Validate;
 use SRS::EPP::Session;
 use Event;
-
 use Log::Log4perl qw(:easy);
-
 use POSIX ":sys_wait_h";
 
 with 'SRS::EPP::Proxy::SimpleConfig';
@@ -24,10 +21,13 @@ with 'MooseX::Log::Log4perl::Easy';
 with 'MooseX::Daemonize';
 
 has '+configfile' => (
-	default => sub { [
-		"$ENV{HOME}/.srs_epp_proxy.yaml",
-		'/etc/srs-epp-proxy.yaml'
-	       ] });
+	default => sub {
+		[
+			"$ENV{HOME}/.srs_epp_proxy.yaml",
+			'/etc/srs-epp-proxy.yaml'
+		];
+		}
+);
 
 sub BUILD {
 	my $self = shift;
@@ -42,25 +42,26 @@ sub BUILD {
 		$logging = "INFO";
 	}
 
-	if ( !ref $logging and ! -f $logging ) {
+	if ( !ref $logging and !-f $logging ) {
+
 		# 'default'
 		if ( $self->is_daemon ) {
 			$logging = {
-		rootLogger => "$logging, Syslog",
-		"appender.Syslog" => "Log::Log4perl::JavaMap::SyslogAppender",
-		"appender.Syslog.logopt" => "pid",
-		"appender.Syslog.Facility" => "daemon",
-		"appender.Syslog.layout" =>
-			"Log::Log4perl::Layout::SimpleLayout",
+				rootLogger => "$logging, Syslog",
+				"appender.Syslog" => "nz.net.nzrs.SyslogAppender",
+				"appender.Syslog.logopt" => "pid",
+				"appender.Syslog.Facility" => "daemon",
+				"appender.Syslog.layout" =>
+					"Log::Log4perl::Layout::SimpleLayout",
 			};
 		}
 		else {
 			$logging = {
-		rootLogger => "$logging, Screen",
-		"appender.Screen" => "Log::Log4perl::Appender::Screen",
-		"appender.Screen.stderr" => 1,
-		"appender.Screen.layout" =>
-			"Log::Log4perl::Layout::SimpleLayout",
+				rootLogger => "$logging, Screen",
+				"appender.Screen" => "Log::Log4perl::Appender::Screen",
+				"appender.Screen.stderr" => 1,
+				"appender.Screen.layout" =>
+					"Log::Log4perl::Layout::SimpleLayout",
 			};
 		}
 	}
@@ -68,16 +69,22 @@ sub BUILD {
 	# prepend "log4perl." to config hashes
 	if ( ref $logging and ref $logging eq "HASH" ) {
 		for my $key ( keys %$logging ) {
-			if ( $key !~ /^log4perl\./ and
-				     !exists $logging->{"log4perl.$key"}
-				    ) {
+			if (
+				$key !~ /^log4perl\./
+				and
+				!exists $logging->{"log4perl.$key"}
+				)
+			{
 				$logging->{"log4perl.$key"} =
 					delete $logging->{$key};
 			}
 		}
 	}
+	
+    $Log::Log4perl::JavaMap::user_defined{'nz.net.nzrs.SyslogAppender'} = 'SRS::EPP::Logging::SyslogAppender';
+
+	Log::Log4perl->init($logging);
 
-	Log::Log4perl->init( $logging );
 	# pass configuration options to the session class?
 }
 
@@ -98,15 +105,15 @@ has 'listener' =>
 	is => "rw",
 	isa => "SRS::EPP::Proxy::Listener",
 	default => sub {
-		require SRS::EPP::Proxy::Listener;
-		my $self = shift;
-		SRS::EPP::Proxy::Listener->new(
-			($self->listen ? (listen => $self->listen) : () ),
-		       );
+	require SRS::EPP::Proxy::Listener;
+	my $self = shift;
+	SRS::EPP::Proxy::Listener->new(
+		($self->listen ? (listen => $self->listen) : () ),
+	);
 	},
 	lazy => 1,
 	handles => {
-		'init_listener' => 'init',
+	'init_listener' => 'init',
 	},
 	;
 
@@ -136,15 +143,16 @@ has 'server_name' =>
 	isa => "Str",
 	lazy => 1,
 	default => sub {
-		my $self = shift;
-		my @listen = @{ $self->listen };
-		if ( @listen == 1 and $listen[0] !~ /^(?:\d+\.|\[)/ ) {
-			# listen address seems a reasonable default...
-			$listen[0];
-		}
-		else {
-			hostname;
-		}
+	my $self = shift;
+	my @listen = @{ $self->listen };
+	if ( @listen == 1 and $listen[0] !~ /^(?:\d+\.|\[)/ ) {
+
+		# listen address seems a reasonable default...
+		$listen[0];
+	}
+	else {
+		hostname;
+	}
 	};
 
 has 'ssl_engine' =>
@@ -162,17 +170,21 @@ use Net::SSLeay::OO;
 use Net::SSLeay::OO::Error qw(die_if_ssl_error);
 use Net::SSLeay::OO::Constants
 	qw(MODE_ENABLE_PARTIAL_WRITE MODE_ACCEPT_MOVING_WRITE_BUFFER
-	   OP_ALL OP_NO_SSLv2 VERIFY_PEER VERIFY_FAIL_IF_NO_PEER_CERT
-	   FILETYPE_PEM);
+	OP_ALL OP_NO_SSLv2 VERIFY_PEER VERIFY_FAIL_IF_NO_PEER_CERT
+	FILETYPE_PEM);
 
-method init_ssl() {
-	my $ctx = Net::SSLeay::OO::Context->new;
+sub init_ssl {
+	my ($self) = @_;
+
+	my $ctx = Net::SSLeay::OO::Context->new(
+		use_default_verify_paths => 0,
+	);
 	$ctx->set_options(&OP_ALL | OP_NO_SSLv2);
 	my $options = VERIFY_PEER;
 	if ( $self->rfc_compliant_ssl) {
 		$self->log_info(
-"Strict RFC5734-compliant SSL enabled (client certificates required)"
-		       );
+			"Strict RFC5734-compliant SSL enabled (client certificates required)"
+		);
 		$options |= VERIFY_FAIL_IF_NO_PEER_CERT;
 	}
 	$ctx->set_verify($options);
@@ -180,19 +192,23 @@ method init_ssl() {
 	$ctx->load_verify_locations("", $self->ssl_cert_dir);
 	$self->log_info(
 		"SSL private key: ".$self->ssl_key_file
-		.", public certificate chain: ".$self->ssl_cert_file
-	       );
+			.", public certificate chain: ".$self->ssl_cert_file
+	);
 	$ctx->use_PrivateKey_file($self->ssl_key_file, FILETYPE_PEM);
 	$ctx->use_certificate_chain_file($self->ssl_cert_file);
 	die_if_ssl_error;  # one last check...
 	$self->ssl_engine($ctx);
 }
 
-method init() {
+sub init {
+    my ($self) = @_;
+
 	$self->log_info("Initializing PGP");
 	$self->init_pgp;
 	$self->log_info("Initializing SSL");
 	$self->init_ssl;
+	$self->log_info("Initializing URIs");
+	$self->init_uris;
 	$self->log_info("Initializing Listener");
 	$self->init_listener;
 }
@@ -202,17 +218,18 @@ has 'openpgp' =>
 	isa => "SRS::EPP::OpenPGP",
 	lazy => 1,
 	default => sub {
-		my $self = shift;
-		require SRS::EPP::OpenPGP;
-		my $pgp_dir = $self->pgp_dir;
-		my $secring_file = "$pgp_dir/secring.gpg";
-		my $pubring_file = "$pgp_dir/pubring.gpg";
-		my $pgp = SRS::EPP::OpenPGP->new(
-			public_keyring => $pubring_file,
-			secret_keyring => $secring_file,
-		       );
-		$pgp->uid($self->pgp_keyid);
-		$pgp;
+	my $self = shift;
+	require SRS::EPP::OpenPGP;
+	my $pgp_dir = $self->pgp_dir;
+	my $secring_file = "$pgp_dir/secring.gpg";
+	my $pubring_file = "$pgp_dir/pubring.gpg";
+	my $pgp = SRS::EPP::OpenPGP->new(
+		public_keyring => $pubring_file,
+		secret_keyring => $secring_file,
+	);
+	$pgp->uid($self->pgp_keyid) if $self->pgp_keyid;
+	my $key = $pgp->default_signing_key;
+	$pgp;
 	},
 	handles => ["pgp"],
 	;
@@ -221,21 +238,52 @@ has 'pgp_keyid' =>
 	metaclass => "Getopt",
 	is => "ro",
 	isa => "Str",
-	required => 1,
 	;
 
 has 'pgp_dir' =>
 	is => "ro",
 	isa => "Str",
 	default => sub {
-		$ENV{GNUPGHOME} || "$ENV{HOME}/.gnupg";
+	$ENV{GNUPGHOME} || "$ENV{HOME}/.gnupg";
 	},
 	;
 
-method init_pgp() {
+sub init_pgp {
+    my ($self) = @_;
+
 	$self->pgp;
 }
 
+has 'extensions' =>
+	metaclass => "Getopt",
+	is => "ro",
+	isa => "HashRef",
+	required => 0,
+	;
+	
+has 'services' =>
+	metaclass => "Getopt",
+	is => "ro",
+	isa => "ArrayRef",
+	required => 1,
+	;
+
+sub init_uris {
+    my ($self) = @_;
+
+	# Register namespaces to be returned by greeting
+	use XML::EPP;
+	XML::EPP::register_obj_uri(
+		@{ $self->services },
+	);
+	
+	if ($self->extensions) {
+		XML::EPP::register_ext_uri(
+			%{ $self->extensions },
+		);
+	}
+}
+
 has 'running' =>
 	is => "rw",
 	isa => "Bool",
@@ -254,10 +302,48 @@ has 'backend' =>
 	default => "https://srstest.srs.net.nz/srs/registrar",
 	;
 
-method accept_one() {
+has 'timeout' =>
+	is => "ro",
+	isa => "Int",
+	default => 300,
+	;
+	
+has 'connection_count' =>
+    is => 'rw',
+    isa => 'Num',
+    default => 0,
+    traits  => ['Counter'],
+    handles => {
+        inc_connection_count   => 'inc',
+        dec_connection_count   => 'dec',
+        reset_connection_count => 'reset',
+    },    
+    ;
+    
+has 'max_connections' =>
+	metaclass => "Getopt",
+	is => "ro",
+	isa => "Int",
+	default => 50,
+	;    
+
+sub accept_one {
+    my ($self) = @_;
+
 	$self->log_trace("accepting connections");
 	my $socket = $self->listener->accept
 		or return;
+		
+    if ($self->connection_count >= $self->max_connections) {
+        $self->log_info("At threshold of " . $self->connection_count . " simultaneous connections. " .
+            "Not responding to connection attempt from " . $socket->peerhost);
+        $socket->print("Server Busy");
+        $socket->close();
+        return;
+    }
+    
+    $self->inc_connection_count;
+    $self->log_debug("Connection begun. Connection count now at: " . $self->connection_count);
 
 	if ( !$self->foreground and (my $pid = fork) ) {
 		push @{ $self->child_pids }, $pid;
@@ -265,21 +351,43 @@ method accept_one() {
 		return ();
 	}
 	else {
+
 		# We'll also want to know the address of the other end
 		# of the socket, for checking it against the back-end
 		# ACL
 		my $peerhost = $socket->peerhost;
 		$self->log_info("connection from $peerhost, starting SSL");
 		$0 = "srs-epp-proxy [$peerhost] - SSL init";
+		
+        my $ssl;
+        eval {
+            # Don't use catch_signal(), as we need to handle it right away
+            local $SIG{ALRM} = sub {
+                die "Timed out waiting for a SSL handshake to complete\n";
+            };
+
+            alarm $self->timeout;
+            $ssl = $self->ssl_engine->accept($socket);
+            alarm 0;
+        };
+
+		my $error = $@;
+		if ($error) {
+
+			# We got an SSL error - send it back to the client, and close the connection
+			$socket->print($error);
+			$socket->close();
+			die $error;
+		}
 
-		my $ssl = $self->ssl_engine->accept($socket);
 		$0 = "srs-epp-proxy [$peerhost] - setup";
 
 		# RFC3734 and updates specify the use of client
 		# certificates.  So, fetch it and get its subject.
 		my $client_cert = $ssl->get_peer_certificate;
 		my $peer_cn;
-		if ( $client_cert ) {
+		if ($client_cert) {
+
 			# should use subjectAltName if present..
 			$peer_cn = $client_cert->get_subject_name->cn;
 			$self->log_info("have a valid peer certificate, cn=$peer_cn");
@@ -289,8 +397,10 @@ method accept_one() {
 		}
 
 		# set the socket to non-blocking for event-driven fun.
-		my $mode = ( MODE_ENABLE_PARTIAL_WRITE |
-				     MODE_ACCEPT_MOVING_WRITE_BUFFER );
+		my $mode = (
+			MODE_ENABLE_PARTIAL_WRITE |
+				MODE_ACCEPT_MOVING_WRITE_BUFFER
+		);
 		$ssl->set_mode($mode);
 		$socket->blocking(0);
 
@@ -299,11 +409,13 @@ method accept_one() {
 			io => $ssl,
 			proxy => $self,
 			socket => $socket,
+			($self->timeout ? (timeout => $self->timeout) : ()),
 			backend_url => $self->backend,
 			event => "Event",
 			peerhost => $peerhost,
-			($self->rfc_compliant_ssl ? (peer_cn => $peer_cn) : ()),
-		       );
+			($self->rfc_compliant_ssl ? (peer_cn => lc $peer_cn) : ()),
+		);
+
 		# let it know it's connected.
 		$session->connected;
 
@@ -311,9 +423,16 @@ method accept_one() {
 	}
 }
 
-method show_state( Str $state, SRS::EPP::Session $session? ) {
+sub show_state {
+    my $self = shift;
+    my ($state,$session) = pos_validated_list(
+        \@_,
+        { isa => 'Str' }, 
+        { isa => 'SRS::EPP::Session', optional => 1 },
+    );
+
 	my ($regid, $peer_host_or_cn);
-	if ( $session ) {
+	if ($session) {
 		$regid = $session->user;
 		$peer_host_or_cn = $session->peer_cn
 			|| $session->peerhost;
@@ -334,12 +453,16 @@ has handlers =>
 	default => sub { {} },
 	;
 
-method signal_handler( Str $signal ) {
+sub signal_handler {
+    my ($self,$signal) = @_;
+
 	$self->log_debug("caught SIG$signal");
 	$self->signals->{$signal}++;
 }
 
-method process_signals() {
+sub process_signals {
+    my ($self) = @_;
+
 	my $sig_h = $self->signals;
 	while (my ($signal,$handler) = each %{ $self->handlers }) {
 		if ($sig_h->{$signal}) {
@@ -350,39 +473,77 @@ method process_signals() {
 	}
 }
 
-method catch_signal(Str $sig, CodeRef $sub) {
+sub catch_signal {
+    my $self = shift;
+    my ($sig,$sub) = pos_validated_list(
+        \@_,
+        { isa => 'Str' }, 
+        { isa => 'CodeRef' }
+    );
+
 	$self->handlers->{$sig} = $sub;
 	$SIG{$sig} = sub { $self->signal_handler($sig) };
 }
 
-method accept_loop() {
-	$self->catch_signal(TERM => sub {
-				    $self->log_info("Shutting down.");
-				    for my $kid ( @{ $self->child_pids } ) {
-					    kill "TERM", $kid;
-				    }
-				    $self->running(0);
-			    });
+sub accept_loop {
+    my ($self) = @_;
+
+    # Setup sig handle for INT/TERM
+    #  (note, shutting down via MooseX::Daemonize normally uses INT)
+    my $sig_handler = sub {
+    	$self->log_info("Shutting down.");
+    	for my $kid ( @{ $self->child_pids } ) {
+    		kill "TERM", $kid;
+    	}
+    	$self->running(0);
+    };
+    
+    for my $sig (qw(INT TERM)) {
+    	$self->catch_signal(
+    		$sig => $sig_handler, 
+    	);
+    }
+	
 	if ( !$self->foreground ) {
-		$self->catch_signal(CHLD => sub { $self->reap_children });
+		$self->catch_signal(CHLD => sub { 
+		    $self->reap_children;
+	    });
 	}
 	$0 = "srs-epp-proxy - listener";
 	while ( $self->running ) {
-		my $session = $self->accept_one;
-		if ( $session ) {
+		my $session = eval {
+		    $self->accept_one;
+		};
+		my $error = $@;
+		if ($error) {
+            $self->log_error("Got error while accepting connection: $error");  
+            exit unless $self->foreground;
+		}
+		
+		if ($session) {
 			unless ( $self->foreground ) {
-				$self->catch_signal(TERM => sub {
-							    $session->shutdown;
-						    });
+				$self->catch_signal(
+					TERM => sub {
+						$session->shutdown;
+					}
+				);
 			}
 			$self->log_trace("accepted a new session, entering event loop");
+
+			# if an untrapped error occurs, we have to just bail out.
+			# events need to trap errors themselves to avoid this.
 			local($Event::DIED) = sub {
 				my $event = shift;
 				my $exception = shift;
-				$self->log_error("Exception during ".$event->w->desc."; $exception");
+				my $estr = "Unhandled exception during ".$event->w->desc."; $exception";
+				if ( open(my $out,">","/tmp/proxy-crashlog") ) {
+					# Just in case the logging isn't working....
+					print $out "$estr\n";
+				}
+				$self->log_error($estr);
 				Event::unloop_all;
 			};
-			Event::loop(120);
+			Event::loop();
 			$self->log_info("Session ends");
 			exit unless $self->foreground;
 		}
@@ -393,18 +554,27 @@ method accept_loop() {
 	}
 }
 
-method reap_children() {
+sub reap_children {
+	my ($self) = @_;
 	my $kid;
 	my %reaped;
 	do {
 		$kid = waitpid(-1, WNOHANG);
 		if ($kid > 0) {
 			$reaped{$kid} = $?;
-			$self->log_debug(
-"child $kid, ".($?&255 ?" killed by signal ".($?&127)
-			.($?&128?" (core dumped)":"")
-:"exited with error code ".($?>>8))
-			       );
+			$self->log_info(
+				"child $kid, ".(
+					$?&255
+					?" killed by signal "
+						.($?&127)
+						.($?&128?" (core dumped)":"")
+					:"exited with error code ".($?>>8)
+					)
+			);
+
+			$self->dec_connection_count;
+			$self->log_debug("Connection ended. Connection count now at: " . $self->connection_count); 			
+			
 		}
 	} while ($kid > 0);
 	my $child_pids = $self->child_pids;
@@ -419,11 +589,13 @@ method reap_children() {
 		my %args = @_;
 		$args{dont_close_all_files} = 1;
 		$SIG{__DIE__} = sub {
+
 			# be sure to re-throw exceptions whilst inside
 			# eval { }
-			if ( $^S ) {
+			if ($^S) {
 				die @_;
-			} else {
+			}
+			else {
 				$self->log_error("Uncaught exception, exiting: @_");
 				$self->log_error("stack trace: ".Carp::longmess);
 				exit(1);
@@ -447,10 +619,36 @@ before 'start' => sub {
 
 after 'start' => sub {
 	my $self = shift;
-	$self->accept_loop
-		if $self->is_daemon;
+	if ($self->is_daemon) {
+		$self->log_info("Dropping Privileges");
+		$self->drop_privs;
+		$self->accept_loop;
+	}
 };
 
+has 'user' =>
+	is => "ro",
+	isa => "Str",
+	default => "nobody",
+	;
+
+sub drop_privs {
+    my ($self) = @_;
+
+	if ( $< and $> ) {
+		$self->log_info("Not dropping privilegs, already UID $<");
+	}
+	my $user = $self->user;
+	my ($uid, $gid) = (getpwnam $user)[2,3] or do {
+		$self->log_error("cannot drop privileges; no such user '$user'");
+	};
+	my $group = getgrgid($gid) || $gid;
+	$self->log_debug("Setting UID:GID to $user:$group");
+
+	$( = $) = $gid;
+	$< = $> = $uid;
+}
+
 1;
 
 __END__
