@@ -1,10 +1,12 @@
 
 package SRS::EPP::Proxy::Listener;
+{
+  $SRS::EPP::Proxy::Listener::VERSION = '0.22';
+}
 
 use 5.010;  # for (?| alternation feature
 
 use Moose;
-use MooseX::Method::Signatures;
 
 with 'MooseX::Log::Log4perl::Easy';
 
@@ -12,19 +14,20 @@ use IO::Select;
 use Net::SSLeay::OO;
 use Socket;
 use IO::Socket::INET;
+use MooseX::Params::Validate;
 
 our ($HAVE_V6, @SOCKET_TYPES);
+
 BEGIN {
 	my $sock = eval {
-		require Socket6;
-		Socket6->import;
+    use if $] < 5.014, "Socket6";
 		require IO::Socket::INET6;
 		IO::Socket::INET6->new(
 			Listen    => 1,
 			LocalAddr => '::1',
 			LocalPort => int(rand(60000)+1024),
 			Proto     => 'tcp',
-		       );
+		);
 	};
 	if ( $sock or $!{EADDRINUSE} ) {
 		$HAVE_V6 = 1;
@@ -37,13 +40,19 @@ sub resolve {
 	my $hostname = shift;
 	my @addr;
 	$DB::single = 1;
-	if ( $HAVE_V6 ) {
+	if ($HAVE_V6) {
 		my @res = getaddrinfo($hostname, "", AF_UNSPEC);
-		while ( my ($family, $socktype, $proto, $address,
-			    $canonical) = splice @res, 0, 5 ) {
+		while (
+			my (
+				$family, $socktype, $proto, $address,
+				$canonical
+			)
+			= splice @res, 0, 5
+			)
+		{
 			my ($addr) = getnameinfo($address, &NI_NUMERICHOST);
 			push @addr, $addr unless grep { $_ eq $addr }
-				@addr;
+					@addr;
 		}
 	}
 	else {
@@ -82,7 +91,8 @@ sub fmt_addr_port {
 	}
 }
 
-method init() {
+sub init {
+    my $self = shift;
 
 	my @sockets;
 	for my $addr ( @{ $self->listen } ) {
@@ -91,13 +101,13 @@ method init() {
 		# way to supply a default port number.
 		my ($hostname, $port) = $addr =~
 			m{^(?|\[([^]]+)\]|([^:]+))(?::(\d+))?$}
-				or die "bad listen address: $addr";
+			or die "bad listen address: $addr";
 		$port ||= EPP_DEFAULT_LOCAL_PORT;
 
 		my @addr = resolve($hostname);
 		$self->log_debug("$hostname resolved to @addr");
 
-		for my $addr ( @addr ) {
+		for my $addr (@addr) {
 			my $SOCKET_TYPE = "IO::Socket::INET";
 			if ( $addr =~ /:/ ) {
 				$SOCKET_TYPE .= "6";
@@ -108,19 +118,19 @@ method init() {
 				LocalPort => $port,
 				Proto => "tcp",
 				ReuseAddr => 1,
-			       );
+			);
 
 			my $addr_port = fmt_addr_port($addr,$port);
 
 			if ( !$socket ) {
 				$self->log_error(
-				"Failed to listen on $addr_port; $!",
-				       );
+					"Failed to listen on $addr_port; $!",
+				);
 			}
 			else {
 				$self->log_info(
 					"Listening on $addr_port",
-				       );
+				);
 				push @sockets, $socket;
 			}
 		}
@@ -133,10 +143,17 @@ method init() {
 	@{ $self->sockets } = @sockets;
 }
 
-method accept( Int $timeout? ) {
+sub accept {
+    my $self = shift;
+    
+    my ( $timeout ) = pos_validated_list(
+        \@_,
+        { isa => 'Int', optional => 1 },
+    );        
+    
 	my $select = IO::Select->new();
 	$select->add($_) for @{$self->sockets};
-	my @ready = $select->can_read( $timeout )
+	my @ready = $select->can_read($timeout)
 		or return;
 	while ( @ready > 1 ) {
 		if ( rand(1) > 0.5 ) {
@@ -153,7 +170,9 @@ method accept( Int $timeout? ) {
 	$socket;
 }
 
-method close() {
+sub close {
+    my $self = shift;
+    
 	for my $socket ( @{ $self->sockets } ) {
 		$socket->close if $socket;
 	}
