@@ -1,15 +1,17 @@
 
 package SRS::EPP::Request;
+{
+  $SRS::EPP::Request::VERSION = '0.22';
+}
 
 use Moose;
-use Moose::Util::TypeConstraints;
 
 extends 'SRS::EPP::Message';
 
 use XML::SRS;
 has "+message" =>
 	isa => "XML::SRS::Action|XML::SRS::Query",
-	handles => [ qw(action_id) ],
+	handles => [qw(action_id)],
 	;
 
 1;
