@@ -10,12 +10,15 @@ use strict;
 use warnings;
 
 package SRS::EPP::Response;
+{
+  $SRS::EPP::Response::VERSION = '0.22';
+}
 use Moose;
-use Moose::Util::TypeConstraints;
-use MooseX::Method::Signatures;
+use Moose::Util::TypeConstraints qw(subtype coerce as where enum);
 extends 'SRS::EPP::Message';
 
 use SRS::EPP::SRSResponse;
+use XML::EPP::Extension;
 
 has 'code' =>
 	is => 'ro',
@@ -31,48 +34,69 @@ has 'extra' =>
 has 'msgQ' =>
 	is => "ro",
 	isa => "XML::EPP::msgQType";
-	;
 
 has 'payload' =>
 	is => "ro",
 	;
 
+has 'extension' =>
+	is => "ro",
+	;
+
 use XML::EPP;
 has "+message" =>
 	isa => "XML::EPP",
+	lazy => 1,
 	default => sub {
-		my $self = shift;
-		my $server_id = $self->server_id;
-		my $client_id = $self->client_id;
-		my $tx_id;
-		if ( $server_id ) {
-			$tx_id = XML::EPP::TrID->new(
-				server_id => $server_id,
-				($client_id ? (client_id => $client_id) : () ),
-				);
-		}
-		my $msg = $self->extra;
-		my $result = XML::EPP::Result->new(
-			($msg ? (msg => $msg) : ()),
-			code => $self->code,
-		       );
-		my $payload;
-		if ( $self->payload ) {
-			$payload = XML::EPP::SubResponse->new(
-				payload => $self->payload,
-			);
-		}
-		XML::EPP->new(
-			message => XML::EPP::Response->new(
-				result => [ $result ],
-				($payload ? (response => $payload) : ()),
-				($self->msgQ ? (msgQ => $self->msgQ) : ()),
-				($tx_id ? (tx_id => $tx_id) : () ),
-			),
-		);
+	my $self = shift;
+	$self->build_response;
 	},
 	;
 
+sub build_response {
+    my $self = shift;
+    
+	my $server_id = $self->server_id;
+	my $client_id = $self->client_id;
+	my $tx_id;
+	if ($server_id) {
+		$tx_id = XML::EPP::TrID->new(
+			server_id => $server_id,
+			($client_id ? (client_id => $client_id) : () ),
+		);
+	}
+	my $msg = $self->extra;
+	my $result = XML::EPP::Result->new(
+		($msg ? (msg => $msg) : ()),
+		code => $self->code,
+	);
+	my ($payload, $extension);
+	if ( $self->payload ) {
+		$payload = XML::EPP::SubResponse->new(
+			payload => $self->payload,
+		);
+	}
+	
+	if ($self->extension) {
+		# We only support one extension at the moment...
+		$extension = XML::EPP::Extension->new(
+			ext_objs => [$self->extension],
+		);
+
+	}
+
+	
+	XML::EPP->new(
+		message => XML::EPP::Response->new(
+			result => [$result],
+			($payload ? (response => $payload) : ()),
+			($extension ? (extension => $extension) : ()),
+			($self->msgQ ? (msgQ => $self->msgQ) : ()),
+			($tx_id ? (tx_id => $tx_id) : () ),
+		),
+	);
+}
+
 has "client_id" =>
 	is => "ro",
 	isa => "XML::EPP::trIDStringType",
@@ -89,6 +113,14 @@ use Module::Pluggable
 	search_path => [__PACKAGE__],
 	;
 
+sub ids {
+	my $self = shift;
+	return (
+		$self->server_id || sprintf("0x%x",(0+$self)),
+		$self->client_id||(),
+	);
+}
+
 __PACKAGE__->plugins;
 
 no Moose;
