@@ -1,5 +1,8 @@
 
 package SRS::EPP::SRSResponse;
+{
+  $SRS::EPP::SRSResponse::VERSION = '0.22';
+}
 
 use Moose;
 
@@ -11,4 +14,9 @@ has "+message" =>
 	handles => [qw(action_id)],
 	;
 
+sub ids {
+	my $self = shift;
+	$self->message->results->[0]->result_id;
+}
+
 1;
